////////////////////////////////////////////////////////////////////
//
// Lacewing networking platform
// http://lacewing.aquadasoft.com
//
// Copyright  Jamie McLaughlin, 2007-2010
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
// FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL JAMIE MCLAUGHLIN
// OR AQUADASOFT BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
// AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. THE SOFTWARE
// MAY BE USED IN EITHER COMMERCIAL AND NON-COMMERCIAL WORKS, PROVIDED THAT IN BOTH
// CASES CREDIT TO "Jamie McLaughlin" AND THE INTERNET URL "http://www.aquadasoft.com"
// ARE VISIBLY INCLUDED IN ANY CREDIT SCREEN AND/OR SPLASH SCREEN, AND ANY DOCUMENTATION,
// WITHOUT ANY SUGGESTION THAT JAMIE MCLAUGHLIN OR AQUADASOFT SUPPORT OR ENDORSE THE
// WORK IN QUESTION. YOU MAY NOT MODIFY, DISASSEMBLE OR DECOMPILE THE SOFTWARE IN WHOLE
// OR IN PART.
//
// Contact: jamie@aquadasoft.com
// Website: http://www.aquadasoft.com
//
// Special thanks to Mathias Krlev, Jeff Vance, Chris Branch, Lukas Meller, Mark Hendriks and Jean Villy Edberg
// This file is designed to be viewed in an IDE like MSVC - please don't use notepad, it screws up the spacing
//
// If you want Lacewing classes to work with STL string, make sure you #include <string> before Lacewing.h
// If you want Lacewing classes to work with MFC CString, make sure you #include <afxwin.h> before Lacewing.h
//
// Utilities:
//
// Lacewing::Sync - synchronise access to data between different threads
// Lacewing::Event - trigger events in one thread from another
// Lacewing::Ping - send ICMP ping requests to a hostname or IP address
// Lacewing::LAN - enumerate devices on a local area network
//
// Sockets:
//
// Lacewing::SocketClient - highly optimised, multithreaded socket client
// Lacewing::SocketServer - highly optimised, multithreaded socket server
// Lacewing::UDP - fast, connectionless communication using UDP rather than TCP
//
// Protocol:
//
// Lacewing::Client - complete implementation of the Lacewing protocol (client-side)
// Lacewing::Server - complete implementation of the Lacewing protocol (server-side)
// Lacewing::Webserver - host a simple webserver, controlled by your C++ code
//
////////////////////////////////////////////////////////////////////



// Include guard
#ifndef LacewingIncluded
#define LacewingIncluded

// Check for a C++ compiler
#ifndef __cplusplus
#pragma error("Lacewing is not compatible with C - please use a C++ compiler")
#else

// Define LacewingFunction to import functions from the DLL
#ifndef LacewingFunction
#define LacewingFunction __declspec(dllimport)

// Automatically link with Lacewing.lib if Lacewing_NoAutoLink isn't defined
#ifndef Lacewing_NoAutoLink
#pragma comment(lib,"Lacewing.lib")
#endif
#endif

// Lacewing namespace
namespace Lacewing
{

// The Event class for the event functions
class Event;

///////////////////////////////
///
///  Global functions
///
//////////////////////////////

// Retrieve the date on which this version of the Lacewing library was compiled
LacewingFunction const char * Version();

// Initialise Lacewing. Set MFC to true to use AfxBeginThread for multithreading.
LacewingFunction bool Initialise(bool MFC = false);

// Free Lacewing
LacewingFunction void Free();

// Start a thread
LacewingFunction bool StartThread(void * Function, void * Parameter=0);

// Pause the thread for a number of milliseconds
LacewingFunction void Pause(int Milliseconds);

// Wait for a single event to trigger. Returns true if the event signaled or false if the timeout elapsed.
LacewingFunction bool WaitEvent(Event &Event, int Timeout=-1);

// Wait for multiple events to trigger. Returns true if the event signaled or false if the timeout elapsed.
LacewingFunction bool WaitEvents(bool All, int Timeout, ...);

// Resolve an string IP address or hostname into a 4 byte address (long integer)
LacewingFunction int Resolve(const char * Address);

// Resolve an string IP address or hostname into a 4 byte address (long integer) (unicode)
LacewingFunction int Resolve(const short * Address);

// Pause the current thread until the console is closed
LacewingFunction void SuspendConsole();


///////////////////////////////
///
///  Enumerations
///
//////////////////////////////

// Transmission modes

// These specify how data is to be (or was) transmitted. Send uses TCP (guaranteed but slow) and
// blast uses UDP (non-guaranteed but fast)

namespace TransmissionMode
{
	enum Type
	{
		Send						= 0,
		Blast						= 1,
	};

	LacewingFunction const char * ToString(Type);
};

// Control

// These are returned from the handlers called by Lacewing to specify how behaviour
// should be controlled. Lacewing::Control::Cancel will cancel the default behaviour,
// and Lacewing::Control::Passive will have no effect on the default behaviour.

namespace Control
{
	enum Type
	{
		Cancel						= 0,
		Passive						= 1,
	};

	LacewingFunction const char * ToString(Type);
};

// Execution modes

// Some classes have a ::SetMode function, which uses these to set the class into
// multithreaded mode (for thread-safe applications with no heartbeat to call Tick())
// or tick mode (for singlethreaded applications with a heartbeat). It is usually best
// to use the client classes in Tick mode where possible, but thread safety can be
// achieved using HandlerMode::Queue and/or HandlerMode::Suspend in multithreaded mode

namespace ExecutionMode
{
	enum Type
	{
		Multithreaded				= 0,
		Tick						= 1,
	};

	LacewingFunction const char * ToString(Type);
};

// Handler modes

// A handler registered with Lacewing::HandlerMode::Trigger will be called immediately
// by the thread in which the event occurs. This has the least impact on performance
// when allowing handlers to control behaviour, but the handler could be operating in any
// thread, so synchronisation objects may have to be used (see Lacewing::Sync)

// A handler registered with Lacewing::HandlerMode::Queue will be added to a queue of
// handlers when it occurs, and then the handler will be called when ProcessHandlers()
// is called. This has the least impact on performance while remaining thread-safe,
// but handlers have no way of controlling behaviour (as they will be called too late)

// A handler registered with Lacewing::HandlerMode::Suspend will suspend the thread until
// the handler is called by ProcessHandlers(). This has a large impact on performance,
// but is thread-safe and allows handlers to control behaviour. If a handler is ready
// to be called in the same thread it was registered in, it will trigger immediately
// rather than wait for ProcessHandlers()

namespace HandlerMode
{
	enum Type
	{
		Trigger						= 0,
		Queue						= 1,
		Suspend						= 2,
	};

	LacewingFunction const char * ToString(Type);
}

// Reasons a client may be denied from an Lacewing server

namespace DenyStatus
{
	enum Type
	{
		NoReasonGiven				= 0,
		ServerFull					= 1,
		IPBanned					= 2,
		InvalidName					= 3,
		InvalidChannel				= 4,
		UnsupportedVersion			= 6,
		NoPingResponse				= 7,
		NameAlreadyTaken			= 8,
		ProtocolFailure				= 9,
		NotDenied					= 10,
	};

	LacewingFunction const char * ToString(Type);
}

// Error codes

namespace ErrorCode
{

	enum Type
	{
		ErrorBindingPort			= 20001,
		ErrorListening				= 20002,
		ErrorCreatingSocket			= 20003,
		ErrorCreatingEvent			= 20004,
		ErrorSelectingEvent			= 20005,
		ErrorReceiving				= 20006,
		ErrorAccepting				= 20007,
		ErrorSending				= 20008,
		ErrorAddingClient			= 20009,
		ErrorResolvingHostname		= 20010,
		ErrorConnecting				= 20011,
		NoChannel					= 20013,
		NoName						= 20014,
		AlreadyConnected			= 20015,
		ErrorConnectionTime			= 20016,
		IncorrectLoopUsage			= 20017,
		InvalidPort					= 20018,
		ErrorReadingAddress			= 20019,
		NotConnected				= 20020,
		ProxyRejectedConnection		= 20021,
		ProxyRequiresIdentd			= 20022,
		ProxyIdentdFailed			= 20023,
		PortAlreadyTaken			= 20025,
		NoHostSet					= 20026,
		UDPHostListenConflict		= 20027,
		PacketTooLarge				= 20028,
		ErrorCreatingICMP			= 20029,
		ErrorEnumeratingDevices		= 20030,
		ErrorReceivingFile			= 20031,
		ErrorSendingFile			= 20032,
		FileNotFound				= 20037,
		AlreadyListening			= 20038,
		ProtocolFailure				= 20039,
		UnsupportedOperatingSystem	= 20040,

		Interrupted					= 10004,
		InvalidFileHandle			= 10009,
		PermissionDenied			= 10013,
		BadAddress					= 10014,
		InvalidArgument				= 10022,
		TooManySockets				= 10024,
		WouldBlock					= 10035,
		InProgress					= 10036,
		OperationAlreadyInProgress	= 10037,
		NotSocket					= 10038,
		NoDestination				= 10039,
		TooLong						= 10040,
		WrongProtocol				= 10041,
		BadOption					= 10042,
		UnsupportedProtocol			= 10043,
		UnsupportedSocket			= 10044,
		UnsupportedOperation		= 10045,
		UnsupportedProtocolFamily	= 10046,
		UnsupportedAddressFamily	= 10047,
		PortAlreadyUsed				= 10048,
		InvalidAddress				= 10049,
		NetworkDown					= 10050,
		NetworkUnreachable			= 10051,
		NetworkReset				= 10052,
		ConnectionAborted			= 10053,
		ConnectionResetByPeer		= 10054,
		NoBufferSpaceAvailable		= 10055,
		SocketAlreadyConnected		= 10056,
		SocketNotConnected			= 10057,
		SocketShutdown				= 10058,
		TooManyReferences			= 10059,
		ConnectionTimedOut			= 10060,
		ConnectionRefused			= 10061,
		CannotTranslateName			= 10062,
		NameTooLong					= 10063,
		HostDown					= 10064,
		NoRouteToHost				= 10065,
		DirectoryNotEmpty			= 10066,
		TooManyProcesses			= 10067,
		UserQuotaExceeded			= 10068,
		DiskQuotaExceeded			= 10069,
		StaleFileHandleReference	= 10070,
		ItemRemote					= 10071,
		NetworkSubsystemUnavailable	= 10091,
		UnsupportedWinsockVersion	= 10092,
		WinsockNotInitialised		= 10093,
		GracefulShutdownInProgress	= 10101,
		NoMoreResults				= 10102,
		CallCanceled				= 10103,
		InvalidProcedureCallTable	= 10104,
		InvalidServiceProvider		= 10105,
		ProviderLoadFailed			= 10106,
		SystemCallFailed			= 10107,
		ServiceNotFound				= 10108,
		ClassTypeNotFound			= 10109,
		NoMoreLookupServiceResults	= 10110,
		LookupServiceEndCalled		= 10111,
		DatabaseQueryRefused		= 10112,
		HostNotFound				= 11001,
		TryHostLookupAgain			= 11002,
		HostLookupFailed			= 11003,
		NoDataRecord				= 11004,
		// FileNotFound				= 21011,
		Busy						= 21034,
		SocketError					= 21035,
		UDPError					= 21036,
		WebserverError				= 21037,

		DebugCode0					= 21001,
		DebugCode1					= 21002,
		DebugCode2					= 21003,
		DebugCode3					= 21004,
		DebugCode4					= 21005,
		DebugCode5					= 21006,
		DebugCode6					= 21007,
		DebugCode7					= 21008,
		DebugCode8					= 21009,
		DebugCode9					= 21010,
		DebugCodeA					= 21033,
		DebugCodeB					= 21012,
		DebugCodeC					= 21013,
		DebugCodeD					= 21014,
		DebugCodeE					= 21015,
		DebugCodeF					= 21016,

		ExtendedDebugCode0			= 21017,
		ExtendedDebugCode1			= 21018,
		ExtendedDebugCode2			= 21019,
		ExtendedDebugCode3			= 21020,
		ExtendedDebugCode4			= 21021,
		ExtendedDebugCode5			= 21022,
		ExtendedDebugCode6			= 21023,
		ExtendedDebugCode7			= 21024,
		ExtendedDebugCode8			= 21025,
		ExtendedDebugCode9			= 21026,
		ExtendedDebugCodeA			= 21027,
		ExtendedDebugCodeB			= 21028,
		ExtendedDebugCodeC			= 21029,
		ExtendedDebugCodeD			= 21030,
		ExtendedDebugCodeE			= 21031,
		ExtendedDebugCodeF			= 21032,
		
		NoError						= 0xFFFFFFFF,

	};

	LacewingFunction const char * ToString(Type);
}

// Proxy types

namespace Proxy
{
	enum Type
	{
		Socks4,

		// Currently not implemented
		Socks4A,
		Socks5,
	};

	LacewingFunction const char * ToString(Type);
};

///////////////////////////////
///
///  Classes
///
//////////////////////////////

class Error
{

private:

	// Internal tag (used inside the library)
	void * InternalTag;

public:

	// Constructor and destructor just wrap Initialise and Free
	inline Error()
	{
		Initialise();
	}

	inline ~Error()
	{
		Free();
	}

	// Initialise the class
	LacewingFunction void Initialise();

	// Deinitialise the class
	LacewingFunction void Free();

	// Loop through the error codes contained in this error (highest level first)
	LacewingFunction int ErrorCodeLoop(int ID = 0);
	LacewingFunction Lacewing::ErrorCode::Type ErrorCodeLoopIndex(int ID);

	/*
		for(int ID = 0; ID; ID = Error.ErrorCodeLoop(ID))
		{
			Lacewing::ErrorCode::Type ErrorCode = Error.ErrorCodeLoopIndex(ID);
			
			// cout << "Error: " << Lacewing::ErrorCode::ToString(ErrorCode) << endl;
		}

	*/

	// Add a new error code to this error (as the highest)
	LacewingFunction void Add(Lacewing::ErrorCode::Type Error);

	// Clone the error - returns a pointer that must be freed
	LacewingFunction Lacewing::Error * Clone();

	// Convert this error (all error codes) to a single, human readable string
	LacewingFunction const char * ToString();

	// Tag (use for whatever you want if you need to)
	void * Tag;
};

class Sync
{

public:

	// Internal tag (used inside the library)
	void * InternalTag;

	// Constructor and destructor just wrap Initialise and Free
	inline Sync()
	{
		Initialise();
	}

	inline ~Sync()
	{
		Free();
	}

	// Initialise the class
	LacewingFunction void Initialise();

	// Deinitialise the class
	LacewingFunction void Free();

	// Class to safely lock/unlock the synchronisation object, for read access ONLY.
	// Do not write to an resource you only own a ReadLock for. Do not create a Lock
	// when you already have a ReadLock. You may create a ReadLock when you already
	// have a Lock.

	struct ReadLock
	{
		void * InternalTag;

		LacewingFunction ReadLock(Lacewing::Sync &Object);
		LacewingFunction ~ReadLock();
	};
	
	// Class to safely lock/unlock the synchronisation object, for read and write access.
	// Do not create a Lock when you already have a ReadLock, unless you pass it as the
	// second parameter to the constructor of Lock. You may create a ReadLock when you
	// already have a Lock.

	struct Lock
	{
		void * InternalTag;

		LacewingFunction Lock(Lacewing::Sync &Object);
		LacewingFunction Lock(Lacewing::Sync &Object, ReadLock &Upgrade);

		LacewingFunction ~Lock();
	};

	// Tag (use for whatever you want if you need to)
	void * Tag;
};


class Event
{
public:

	// Internal tag (used inside the library)
	void * InternalTag;

	// Constructor and destructor just wrap Initialise and Free
	inline Event()
	{
		Initialise();
	}

	inline ~Event()
	{
		Free();
	}

	// Initialise the class
	LacewingFunction void Initialise();

	// Deinitialise the class
	LacewingFunction void Free();

	// Check if the event is signalled or not
	LacewingFunction bool Check();

	// Signal the event
	LacewingFunction bool Signal();

	// Unsignal the event
	LacewingFunction void Unsignal();

	// Wait for the event to become signaled. Returns true if the event signaled or false if the timeout elapsed.
	LacewingFunction bool Wait(int Timeout = -1);

	// Tag (use for whatever you want if you need to)
	void * Tag;

};

class SocketClient
{

private:

	// Internal tag (used inside the library)
	void * InternalTag;

public:

	// Constructor and destructor just wrap Initialise and Free
	inline SocketClient()
	{
		Initialise();
	}

	inline ~SocketClient()
	{
		Free();
	}

	// Initialise the class
	LacewingFunction void Initialise();
	
	// Deinitialise the class
	LacewingFunction void Free();

	// Connect function
	LacewingFunction void Connect(const char * Host, int Port);
	LacewingFunction void Connect(const char * ProxyAddress, int ProxyPort, const char * Host, int Port, Proxy::Type ProxyType, const char * ProxyUserID);

	// Disconnect function
	LacewingFunction void Disconnect(bool Block = true);

	// Get the IP address of the server as a string
	LacewingFunction const char * ServerIP();

	// Get the server port the client is connected to
	LacewingFunction int ServerPort();

	// A structure to hold the address information of the server connected to
	struct
	{

		// IP of the server the client is connected to
		unsigned char IP[4];

		// Port the client is connected to the server on
		int Port;

	} Address;

	// Generic send function
	LacewingFunction bool Send(const char * Buffer, int Size=-1);

	// Send a file
	LacewingFunction bool SendFile(const char * Filename);

	// Retrieve how long this client has been connected in seconds
	LacewingFunction int ConnectionTime();

	// Check if the client is connected
	LacewingFunction bool Connected();

	// Check if the client is in the process of trying to connect
	LacewingFunction bool Connecting();

	// Trigger any waiting handlers (for Lacewing::HandlerMode::Queue or Lacewing::HandlerMode::Suspend)
	LacewingFunction void ProcessHandlers();

	// Available handlers for this class
	typedef void (__cdecl * HandlerConnect)		(Lacewing::SocketClient &Client);
	typedef void (__cdecl * HandlerDisconnect)	(Lacewing::SocketClient &Client);
	typedef void (__cdecl * HandlerReceive)		(Lacewing::SocketClient &Client, char * Buffer, int Size);
	typedef void (__cdecl * HandlerError)		(Lacewing::SocketClient &Server, Lacewing::Error &);

	// Register handlers
	LacewingFunction void RegisterHandlerConnect(HandlerConnect, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerDisconnect(HandlerDisconnect, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerReceive(HandlerReceive, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerError(HandlerError, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);

	// Tag (use for whatever you want if you need to)
	void * Tag;
};


class SocketServer
{

private:

	// Internal tag (used inside the library)
	void * InternalTag;

public:

	// Constructor and destructor just wrap Initialise and Free
	inline SocketServer()
	{
		Initialise();
	}

	inline ~SocketServer()
	{
		Free();
	}

	// Initialise the class
	LacewingFunction void Initialise();
	
	// Deinitialise the class
	LacewingFunction void Free();

	// Host function
	LacewingFunction void Host(int Port);

	// Stop hosting
	LacewingFunction void Unhost();

	// Check if the server is hosting
	LacewingFunction bool Hosting();

	// Get the port which the server is hosting on (0 if not hosting)
	LacewingFunction int Port();

	// Retrieve the total number of bytes the server has sent and received
	LacewingFunction __int64 BytesSent();
	LacewingFunction __int64 BytesReceived();

	// Structure to represent a client on the server
	struct Client
	{
		// Internal tag (used inside the library)
		void * InternalTag;

		// The address of the client as a 4 byte IP address
		unsigned char IP[4];

		// The port the server is connected to the client on
		int Port;

		// Convert the IP address to a string
		LacewingFunction const char * StringIP();

		// Generic send function
		LacewingFunction bool Send(const char * Buffer, int Size=-1);

		// File send function
		LacewingFunction bool SendFile(const char * Filename, __int64 Offset=0, __int64 Size=0);

		// Disconnect this client
		LacewingFunction void Disconnect();

		// Disconnect this client from within a client loop. The return value is a new loop ID - Index=Client->Disconnect(Index)
		LacewingFunction int Disconnect(int LoopID);

		// Retrieve how long this client has been connected in seconds
		LacewingFunction int ConnectionTime();

		// Retrieve the total number of bytes this client has been sent and the server has received from it
		LacewingFunction __int64 BytesSent();
		LacewingFunction __int64 BytesReceived();

		// Tag (use for whatever you want if you need to)
		void * Tag;
	};

	// The number of clients connected to the server
	LacewingFunction int ClientCount();

	// Loop through the connected clients
	LacewingFunction int ClientLoop(int ID=0);
	LacewingFunction Client &ClientLoopIndex(int ID);
	
	// Prematurely end the client loop
	LacewingFunction void EndClientLoop(int ID);

	/*

		for(int ID = 0; ID; ID = Server.ClientLoop(ID))
		{
			SocketServer::Client &Client = Server.ClientLoopIndex(ID);
			
			// If leaving the client loop before the end
			Server.EndClientLoop(ID);
		}

	*/

	// Trigger any waiting handlers (for Lacewing::HandlerMode::Queue or Lacewing::HandlerMode::Suspend)
	LacewingFunction void ProcessHandlers();

	// Available handlers for this class
	typedef Lacewing::Control::Type	(__cdecl * HandlerConnect)		(SocketServer &Server, SocketServer::Client &Client);
	typedef void					(__cdecl * HandlerDisconnect)	(SocketServer &Server, SocketServer::Client &Client);
	typedef void					(__cdecl * HandlerReceive)		(SocketServer &Server, SocketServer::Client &Client, char * Buffer,int Size);
	typedef void					(__cdecl * HandlerError)		(SocketServer &Server, Lacewing::Error &);
	
	// Register handlers
	LacewingFunction void RegisterHandlerConnect(HandlerConnect, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerDisconnect(HandlerDisconnect, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerReceive(HandlerReceive, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerError(HandlerError, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);

	// Tag (use for whatever you want if you need to)
	void * Tag;

};

class UDP
{

public:

	// Internal tag (used inside the library)
	void * InternalTag;

	// Constructor and destructor just wrap Initialise and Free
	inline UDP()
	{
		Initialise();
	}

	inline ~UDP()
	{
		Free();
	}

	// Initialise the class
	LacewingFunction void Initialise();

	// Deinitialise the class
	LacewingFunction void Free();

	// Listen on a port for incoming data
	LacewingFunction void Host(int Port);

	// Stop listening for incoming data
	LacewingFunction void Unhost();

	// Get the port which is being hosted on (0 if not hosting)
	LacewingFunction int Port();

	// Address structure to identify the source of received data
	struct Address
	{
		// Internal tag (used inside the library)
		void * InternalTag;

		// Empty constructor (if this is used, Initialise must be called before use or deletion)
		inline Address()
		{

		}

		// Constructor and destructor just wrap Initialise and Free
		inline Address(UDP &UDP)
		{
			Initialise(UDP);
		}

		inline ~Address()
		{
			Free();
		}

		// Initialise the class
		LacewingFunction void Initialise(UDP &UDP);

		// Deinitialise the class
		LacewingFunction void Free();

		// Set the address from a hostname
		LacewingFunction void Set(const char * Hostname,int Port);

		// Set the address from an SocketServer::Client
		LacewingFunction void Set(SocketServer::Client &Host,int Port);

		// Set the address and port from an SocketClient
		LacewingFunction void Set(SocketClient &Host);
		
		// Retrieve the IP address as a string
		LacewingFunction const char * IP_AsString();

		// Retrieve the IP address as a 4 byte integer
		LacewingFunction int IP_AsInteger();

		// Retrieve an individual byte of the IP address
		LacewingFunction unsigned char IP_AsByte(int Index);

		// Retrieve the port
		LacewingFunction int Port();

		// Send a UDP packet to this address
		LacewingFunction void Send(const char * Buffer, int Size=-1);

		// Listen for incoming UDP packets from this address
		// This cannot be used at the same time as Host() in the same UDP object
		// It returns the port which is being used locally by the operating system
		// or 0 on failure
		LacewingFunction int Listen(int Port);

		// Stop listening for incoming UDP packets from this address
		LacewingFunction void StopListening();

		// Tag (use for whatever you want if you need to)
		void * Tag;
	};

	// Trigger any waiting handlers (for Lacewing::HandlerMode::Queue or Lacewing::HandlerMode::Suspend)
	LacewingFunction void ProcessHandlers();

	// Available handlers for this class
	typedef void (__cdecl * HandlerReceive)			(UDP &UDP, Lacewing::UDP::Address &Source, char * Buffer,int Size);
	typedef void (__cdecl * HandlerError)			(UDP &UDP, Lacewing::Error &);
	
	// Register handlers
	LacewingFunction void RegisterHandlerReceive(HandlerReceive, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerError(HandlerError, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);

	// Tag (use for whatever you want if you need to)
	void * Tag;
};

class Webserver
{

private:

	// Internal tag (used inside the library)
	void * InternalTag;

protected:
	
	// The basic socket object
	SocketServer Socket;

public:

	// Constructor and destructor just wrap Initialise and Free
	inline Webserver()
	{
		Initialise();
	}

	inline ~Webserver()
	{
		Free();
	}

	// Initialise the class
	LacewingFunction void Initialise();

	// Deinitialise the class
	LacewingFunction void Free();

	// Start the HTTP server on the specified port
	LacewingFunction void Host(int Port=80);

	// Stop the HTTP server
	LacewingFunction void Unhost();

	// Check if the HTTP server is currently running
	LacewingFunction bool Hosting();

	// Retrieve the total number of bytes the webserver has sent and received
	LacewingFunction __int64 BytesSent();
	LacewingFunction __int64 BytesReceived();

	// Structure to represent a request to the server
	struct Request
	{
		// Internal tag (used inside the library)
		void * InternalTag;

		// Close the socket used to receive the request immediately
		LacewingFunction void Disconnect();

		// Redirect the request to another URL
		LacewingFunction void SetRedirect(const char * URL);

		// Set the mime type for the request
		LacewingFunction void SetMimeType(const char * MimeType);

		// Suggests a mime type for a file or file extension
		LacewingFunction const char * SuggestMimeType(const char * Filename);

		// Equivalent to SetMimeType(SuggestMimeType(Filename)). Returns the mime type used.
		LacewingFunction const char * SetSuggestedMimeType(const char * Filename);

		// Send data in response to the request (after calling BeginResponse)
		LacewingFunction void Send(const char * Buffer,int Size=-1);

		// The same as Send, but sends a file instead
		LacewingFunction void SendFile(const char * Filename);

		// Read a header from the request
		LacewingFunction const char * GetHeader(const char * Name);

		// Set a cookie to be stored by the browser
		LacewingFunction void SetCookie(const char * Name, const char * Value);

		// Get a cookie sent by the browser
		LacewingFunction const char * GetCookie(const char * Name);

		// Ask the browser not to cache this response
		LacewingFunction void DisableCache();

		// Set the charset for this response (such as "utf8")
		LacewingFunction void SetCharset(const char * Charset);

		// Get an item from the POST data by name
		LacewingFunction const char * GetPostItem(const char * Name);

		// Get an item from the GET data by name
		LacewingFunction const char * GetGetItem(const char * Name);

		// Strip the GET data from the URL
		LacewingFunction void StripGet();

		// Enable download resuming for this request (default: disabled)
		LacewingFunction void EnableDownloadResuming();

		// Get the requested range beginning and end (for download resuming)
		// If RangeBeginning() is negative, RangeEnd() will be 0
		// and RangeBeginning() will be an offset from the end of the file
		LacewingFunction __int64 RequestedRangeBeginning();
		LacewingFunction __int64 RequestedRangeEnd();

		// If responding to a range request, use these to tell the browser
		// which range you're sending
		LacewingFunction void SetOutgoingRange(__int64 Begin, __int64 End);

		// Stream operators for sending various types
		Request & operator << (char);
		Request & operator << (unsigned char);
		Request & operator << (short);
		Request & operator << (unsigned short);
		Request & operator << (int);
		Request & operator << (unsigned int);
		Request & operator << (const char *);
		Request & operator << (__int64);
		Request & operator << (float);
		Request & operator << (double);

		// Stream operator for sending an STL string (wraps Send)

		#ifdef _STRING_

		inline Request & operator << (const std::string &S)
		{
			Send(S.c_str(), (int) S.length());
			return *this;
		}

		#endif


		// Stream operator for sending an MFC CString (wraps Send)

		#ifdef _AFX
		#ifdef UNICODE

		inline Request & operator << (const CString &S)
		{
			int Size = WideCharToMultiByte(CP_UTF8, 0, (const unsigned short *)S, -1, 0, 0, 0, 0);
			
			if(Size)
			{
				char * Buffer = new char[Size];
				WideCharToMultiByte(CP_UTF8, 0, (const unsigned short *)S, -1, Buffer, Size, 0, 0);
				Send(Buffer, Size - 1); // Don't include the null terminator
				delete [] Buffer;
			}

			return *this;
		}

		#else

		inline Request & operator << (const CString &S)
		{
			Send((const char *)S, S.GetLength());
			return *this;
		}

		#endif
		#endif

		// If enabled, stream operators will send binary data instead of text. Disabled by default.
		LacewingFunction void SetBinaryStream(bool Enabled);

		// Returns the 4 byte IP address of the request as an integer
		LacewingFunction int IP();

		// Get the filename of file post data (for HandlerPreFilePost or HandlerFilePostChunk)
		const char * GetUploadFilename();

		// Get the MIME type of file post data (for HandlerPreFilePost or HandlerFilePostChunk)
		const char * GetUploadMimeType();

		// Get the name (from the form) of file post data (for HandlerPreFilePost or HandlerFilePostChunk)
		const char * GetUploadName();

		// The URL of the request
		const char * URL;

		// The post data of the request (blank if the request is not a Post request)
		const char * PostData;

		// Tag (use for whatever you want if you need to)
		void * Tag;
	};

	// Trigger any waiting handlers (for Lacewing::HandlerMode::Queue or Lacewing::HandlerMode::Suspend)
	LacewingFunction void ProcessHandlers();

	// Available handlers for this class
	typedef void (__cdecl * HandlerGet)					(Lacewing::Webserver &Webserver, Lacewing::Webserver::Request &Request);
	typedef void (__cdecl * HandlerPreFilePost)			(Lacewing::Webserver &Webserver, Lacewing::Webserver::Request &Request);
	typedef void (__cdecl * HandlerFilePostChunk)		(Lacewing::Webserver &Webserver, Lacewing::Webserver::Request &Request, const char * Buffer, int Size);
	typedef void (__cdecl * HandlerFilePostComplete)	(Lacewing::Webserver &Webserver, Lacewing::Webserver::Request &Request);
	typedef void (__cdecl * HandlerPost)				(Lacewing::Webserver &Webserver, Lacewing::Webserver::Request &Request);
	typedef void (__cdecl * HandlerHead)				(Lacewing::Webserver &Webserver, Lacewing::Webserver::Request &Request);
	typedef void (__cdecl * HandlerError)				(Lacewing::Webserver &Webserver, Lacewing::Error &);

	// These probably won't be required for normal use
	typedef void (__cdecl * HandlerConnect)		(Lacewing::Webserver &Webserver, Lacewing::Webserver::Request &Request);
	typedef void (__cdecl * HandlerDisconnect)	(Lacewing::Webserver &Webserver, Lacewing::Webserver::Request &Request);

	// Register handlers
	LacewingFunction void RegisterHandlerGet(HandlerGet, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerPreFilePost(HandlerPreFilePost, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerFilePostChunk(HandlerFilePostChunk, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerFilePostComplete(HandlerFilePostComplete, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerPost(HandlerPost, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerHead(HandlerHead, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerError(HandlerError, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerConnect(HandlerConnect, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerDisconnect(HandlerConnect, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);

	// Tag (use for whatever you want if you need to)
	void * Tag;
};

	
class Client
{

private:

	// Internal tag (used inside the library)
	void * InternalTag;

protected:

	// The basic socket object
	SocketClient Socket;

	// The UDP object
	UDP UDP;

public:

	// Constructor and destructor just wrap Initialise and Free
	inline Client()
	{
		Initialise();
	}

	inline ~Client()
	{
		Free();
	}

	// Initialise the class
	LacewingFunction void Initialise();

	// Deinitialise the class
	LacewingFunction void Free();

	// Switch the class between multithreaded and tick mode
	LacewingFunction void SetMode(Lacewing::ExecutionMode::Type Mode);

	// Connect function
	LacewingFunction void Connect(const char * Host, int Port=6121, bool EnableUDP=true);

	// Disconnect function
	LacewingFunction void Disconnect(bool Block=true);

	// Get the IP address of the server as a string
	LacewingFunction const char * ServerIP();

	// Get the server port the client is connected to
	LacewingFunction int ServerPort();

	// Get the client's unique ID
	LacewingFunction int GetID();

	// Set the client's name
	LacewingFunction void SetName(const char * Name);

	// Get the client's name (blank if no name is set)
	LacewingFunction const char * Name();

	// Request a list of channels from the server
	LacewingFunction void ListChannels();

	// A channel on the server which has not been joined
	struct ChannelListing
	{
		// The number of peers in the channel
		short PeerCount;

		// The name of the channel
		const char * Name;

	};

	// Loop through all of the non-hidden channels on the server, after using ListChannels()
	LacewingFunction int ChannelListLoop(int ID=0);
	LacewingFunction ChannelListing * ChannelListLoopIndex(int ID);
	LacewingFunction void EndChannelListLoop(int ID);

	/*

		for(int ID = 0; ID; ID = Client.ChannelListLoop(ID))
		{
			Lacewing::Client::ChannelListing &Listing = Client.ChannelListLoopIndex(ID);			

			// If leaving the channel list loop before the end
			Channel.EndChannelListLoop();
		}

	*/


	// Join a channel
	LacewingFunction void Join(const char * Channel,bool Hidden=false);

	// Send a message to the server
	LacewingFunction void SendServer(int Subchannel, const char * Buffer, int Size=-1, int Type=0);

	// Blast a message to the server
	LacewingFunction void BlastServer(int Subchannel, const char * Buffer, int Size=-1, int Type=0);

	// A structure to represent a file being sent to the server
	struct File
	{
		// Internal tag (used inside the library)
		void * InternalTag;

		// The name of the file
		const char * Filename;

		// The percentage complete of the file transfer
		int PercentageComplete();
	};

	// Send a file to the server
	LacewingFunction File * SendFile(int Subchannel, const char * Filename);

	// Check if the client is connected
	LacewingFunction bool Connected();

	// Retrieve how long this client has been connected in seconds
	LacewingFunction int ConnectionTime();

	// The channel structure, for the peer's InChannel function to compile
	struct Channel;

	// Information concerning a joined channel
	struct Channel
	{
		// Internal tag (used inside the library)
		void * InternalTag;

		// Get the channel name
		LacewingFunction const char * Name();

		// Send a message to the channel
		LacewingFunction void Send(int Subchannel, const char * Buffer, int Size=-1, int Type=0);
		
		// Blast a message to the channel
		LacewingFunction void Blast(int Subchannel, const char * Buffer, int Size=-1, int Type=0);

		// A structure to represent a file being sent to the channel
		struct File
		{
			// Internal tag (used inside the library)
			void * InternalTag;

			// The name of the file
			const char * Filename;

			// The percentage complete of the file transfer
			int PercentageComplete();

			// Tag (use for whatever you want if you need to)
			void * Tag;
		};

		// Send a file to the channel
		LacewingFunction File * SendFile(const char * Filename, int Subchannel=0, __int64 Offset=0, unsigned int Size=-1, const char * VirtualFilename="");

		// Another client on a channel
		struct Peer
		{

			// Internal tag (used inside the library)
			void * InternalTag;

			// Numeric ID of the peer
			int ID;

			// Send a message to the peer
			LacewingFunction void Send(int Subchannel, const char * Buffer, int Size=-1, int Type=0);

			// Blast a message to the peer
			LacewingFunction void Blast(int Subchannel, const char * Buffer, int Size=-1, int Type=0);

			// Get the peer name
			LacewingFunction const char * Name();

			// A structure to represent a file being sent to the peer
			struct File
			{

				// Internal tag (used inside the library)
				void * InternalTag;

				// The name of the file
				const char * Filename;

				// The percentage complete of the file transfer
				int PercentageComplete();

				// Tag (use for whatever you want if you need to)
				void * Tag;
			};

			// Send a file to the peer
			LacewingFunction File * SendFile(int Subchannel, const char * Filename);

			// Tag (use for whatever you want if you need to)
			void * Tag;
		};

		// Retrieve the number of peers in the channel
		LacewingFunction int PeerCount();

		// Loop through all the peers in the channel
		LacewingFunction int PeerLoop(int ID=0);
		LacewingFunction Peer &PeerLoopIndex(int ID);

		// Prematurely end the peer loop
		LacewingFunction void EndPeerLoop(int ID);

		/*

			for(int ID = 0; ID; ID = Channel.PeerLoop(ID))
			{
				Client::Channel::Peer &Peer = Channel.PeerLoopIndex(ID);
				
				// If leaving the peer loop before the end
				Channel.EndPeerLoop(ID);
			}

		*/

		// Leave the channel
		LacewingFunction void Leave();

		// Tag (use for whatever you want if you need to)
		void * Tag;
	};

	// The number of channels the client is in
	LacewingFunction int ChannelCount();

	// Loop through all the channels the client is connected to
	LacewingFunction int ChannelLoop(int ID=0);
	LacewingFunction Channel &ChannelLoopIndex(int ID);
	LacewingFunction void EndChannelLoop(int ID);

	/*

		for(int ID = 0; ID; ID = Client.ChannelLoop(ID))
		{
			Lacewing::Client::Channel &Channel = Client.ChannelLoopIndex(ID);			
		
			// If leaving the client loop before the end
			Client.EndChannelLoop(ID);
		}

	*/

	// Find a channel which the client has already joined
	LacewingFunction Channel * FindChannel(const char * Name);
	LacewingFunction Channel * FindChannel(unsigned int ID);

	// Set the hook mode
	LacewingFunction void SetHandlerMode(Lacewing::HandlerMode::Type Mode);
	
	// Trigger any waiting handlers (for Lacewing::HandlerMode::Queue or Lacewing::HandlerMode::Suspend)
	LacewingFunction void ProcessHandlers();

	// Available handlers for this class
	typedef void	(__cdecl * HandlerConnect)									(Lacewing::Client &Client,const char * WelcomeMessage);
	typedef void	(__cdecl * HandlerConnectionDenied)							(Lacewing::Client &Client,Lacewing::DenyStatus::Type Reason);
	typedef void	(__cdecl * HandlerDisconnect)								(Lacewing::Client &Client);
	typedef void	(__cdecl * HandlerServerMessage)							(Lacewing::Client &Client,Lacewing::TransmissionMode::Type Mode,int Subchannel,char * Buffer,int Size,int Type);
	typedef void	(__cdecl * HandlerChannelMessage)							(Lacewing::Client &Client,Lacewing::Client::Channel &Channel,Lacewing::Client::Channel::Peer &Peer,Lacewing::TransmissionMode::Type Mode,int Subchannel,char * Buffer,int Size,int Type);
	typedef void	(__cdecl * HandlerPeerMessage)								(Lacewing::Client &Client,Lacewing::Client::Channel &Channel,Lacewing::Client::Channel::Peer &Peer,Lacewing::TransmissionMode::Type Mode,int Subchannel,char * Buffer,int Size,int Type);
	typedef void	(__cdecl * HandlerError)									(Lacewing::Client &Client,Lacewing::Error &);
	typedef void	(__cdecl * HandlerJoin)										(Lacewing::Client &Client,Lacewing::Client::Channel &Target);
	typedef void	(__cdecl * HandlerJoinDenied)								(Lacewing::Client &Client,Lacewing::DenyStatus::Type Reason);
	typedef void	(__cdecl * HandlerLeave)									(Lacewing::Client &Client,Lacewing::Client::Channel &Target);
	typedef void	(__cdecl * HandlerLeaveDenied)								(Lacewing::Client &Client,Lacewing::Client::Channel &Target,Lacewing::DenyStatus::Type Reason);
	typedef void	(__cdecl * HandlerNameSet)									(Lacewing::Client &Client);
	typedef void	(__cdecl * HandlerNameChanged)								(Lacewing::Client &Client,const char * OldName);
	typedef void	(__cdecl * HandlerNameDenied)								(Lacewing::Client &Client,Lacewing::DenyStatus::Type Reason);
	typedef void	(__cdecl * HandlerPeerConnect)								(Lacewing::Client &Client,Lacewing::Client::Channel &Channel,Lacewing::Client::Channel::Peer &Peer);
	typedef void	(__cdecl * HandlerPeerDisconnect)							(Lacewing::Client &Client,Lacewing::Client::Channel &Channel,Lacewing::Client::Channel::Peer &Peer);
	typedef void	(__cdecl * HandlerPeerChangeName)							(Lacewing::Client &Client,Lacewing::Client::Channel &Channel,Lacewing::Client::Channel::Peer &Peer,const char * OldName);
	typedef void	(__cdecl * HandlerChannelListComplete)						(Lacewing::Client &Client);

	// Register handlers
	LacewingFunction void RegisterHandlerConnect(HandlerConnect, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerConnectionDenied(HandlerConnectionDenied, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerDisconnect(HandlerDisconnect, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerServerMessage(HandlerServerMessage, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerChannelMessage(HandlerChannelMessage, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerPeerMessage(HandlerPeerMessage, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerError(HandlerError, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerJoin(HandlerJoin, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerJoinDenied(HandlerJoinDenied, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerLeave(HandlerLeave, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerLeaveDenied(HandlerLeaveDenied, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);	
	LacewingFunction void RegisterHandlerNameSet(HandlerNameSet, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerNameChanged(HandlerNameChanged, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerNameDenied(HandlerNameDenied, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerPeerConnect(HandlerPeerConnect, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerPeerDisconnect(HandlerPeerDisconnect, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerPeerChangeName(HandlerPeerChangeName, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerChannelListComplete(HandlerChannelListComplete, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);

	// Tag (use for whatever you want if you need to)
	void * Tag;
};

class Server
{

private:

	// Internal tag (used inside the library)
	void * InternalTag;

protected:

	// The basic socket object
	SocketServer Socket;

	// The UDP object
	UDP UDP;

	// The webserver object for the web interface
	Webserver Webserver;

public:


	// Constructor and destructor just wrap Initialise and Free
	inline Server()
	{
		Initialise();
	}

	inline ~Server()
	{
		Free();
	}

	// Initialise the class
	LacewingFunction void Initialise();
	
	// Deinitialise the class
	LacewingFunction void Free();

	// Host function
	LacewingFunction void Host(int Port=6121,bool EnableUDP=true);

	// Stop hosting
	LacewingFunction void Unhost();

	// Check if the server is currently hosting
	LacewingFunction bool Hosting();

	// Get the port which the server is hosting on (0 if not hosting)
	LacewingFunction int Port();

	// Host the web interface
	LacewingFunction void HostWebInterface(int Port=6122,const char * Password="");

	// Stop hosting the web interface
	LacewingFunction void UnhostWebInterface();

	// Check if the web interface is currently being hosted
	LacewingFunction bool HostingWebInterface();

	// Add a custom page to the web interface navigation menu
	LacewingFunction void AddWebInterfaceNavigationPage(const char * Name,const char * Target);

	// Add a spacing to the web interface navigation menu (between pages)
	LacewingFunction void AddWebInterfaceNavigationSpacing();

	// Clear the custom pages from the web interface navigation menu
	LacewingFunction void ClearWebInterfaceNavigationPages();

	// Set the welcome message for the server
	LacewingFunction void SetWelcomeMessage(const char * Message);

	// Enable or disable channel listing (enabled by default)
	LacewingFunction void SetChannelListing(bool Enabled);

	// Set the maximum number of connections from a single IP address
	LacewingFunction void SetMaximumConnectionsPerIP(int Maximum=8);

	// Set the amount (in ms) a client's flood timer should increase by with each connection
	LacewingFunction void SetFloodGain(int Gain=500);

	// Set the value (in ms) a client's flood timer must be equal or greater than to deny a connection
	LacewingFunction void SetFloodTolerance(int Tolerance=1500);

	// Set the value (in ms) a client's flood timer should increase by with each denied connection
	LacewingFunction void SetFloodBlock(int Block=10000);

	// Retrieve the total number of bytes the server has sent and received
	LacewingFunction __int64 BytesSent();
	LacewingFunction __int64 BytesReceived();

	// The client structure so that the channel functions compile
	struct Client;

	// Structure to represent a channel on the server
	struct Channel
	{
		// Internal tag (used inside the library)
		void * InternalTag;

		// Numeric ID of the channel
		int ID;

		// Get the name of the channel
		LacewingFunction const char * Name();

		// The number of clients in the channel
		LacewingFunction int ClientCount();

		// Loop through the clients in the channel
		LacewingFunction int ClientLoop(int ID=0);
		LacewingFunction Client &ClientLoopIndex(int ID);

		// Prematurely end the client loop
		LacewingFunction void EndClientLoop(int ID);

		/*

			for(int ID = 0; ID; ID = Channel.ClientLoop(ID))
			{
				Lacewing::Server::Client &Client = Channel.ClientLoopIndex(ID);
						
				// If leaving the client loop before the end
				Channel.EndClientLoop(ID);
			}

		*/

		// A structure to represent a file being sent to the server
		struct File
		{
			// The name of the file
			const char * Filename;

			// The percentage complete of the file transfer
			int PercentageComplete();
		};

		// Send a file to the channel
		LacewingFunction File * SendFile(const char * Filename, int Subchannel=0, __int64 Offset=0, unsigned int Size=-1, const char * VirtualFilename="");

		// Close the channel
		LacewingFunction void Close();
	};

	// The number of channels on the server
	LacewingFunction int ChannelCount();

	// Loop through all the channels
	LacewingFunction int ChannelLoop(int ID=0);
	LacewingFunction Channel &ChannelLoopIndex(int ID);
	LacewingFunction void EndChannelLoop(int ID);

	/*

		for(int ID = 0; ID; ID = Server.ChannelLoop(ID))
		{
			Lacewing::Server::Channel &Channel = Server.ChannelLoopIndex(ID);
			
			// If leaving the channel loop before the end
			Server.EndChannelLoop(ID);
		}

	*/

	// Structure to represent a client on the server
	struct Client
	{
		// Internal tag (used inside the library)
		void * InternalTag;

		// Numeric ID of the client
		int ID;

		// The address of the client as a 4 byte IP address
		unsigned char IP[4];

		// The port the server is connected to the client on
		int Port;

		// Convert the IP address to a string
		LacewingFunction const char * StringIP();

		// Disconnect this client
		LacewingFunction void Disconnect();

		// Disconnect this client from within a client loop. The return value is a new loop ID - Index=Client->Disconnect(Index)
		LacewingFunction int Disconnect(int LoopID);

		// Retrieve how long this client has been connected in seconds
		LacewingFunction int ConnectionTime();

		// Retrieve the total number of bytes this client has been sent and the server has received from it
		LacewingFunction __int64 BytesSent();
		LacewingFunction __int64 BytesReceived();

		// Send a message to the client
		LacewingFunction void Send(int Subchannel, const char * Buffer, int Size=-1, int Type=0);

		// Blast a message to the client
		LacewingFunction void Blast(int Subchannel, const char * Buffer, int Size=-1, int Type=0);

		// Get the client name
		LacewingFunction const char * Name();

		// The number of channels the client is connected to
		LacewingFunction int ChannelCount();

		// Loop through all the channels the client is connected to
		LacewingFunction int ChannelLoop(int ID=0);
		LacewingFunction Channel &ChannelLoopIndex(int ID);

		// Prematurely end the channel loop
		LacewingFunction void EndChannelLoop(int ID);

		/*

			for(int ID = 0; ID; ID = Client.ChannelLoop(ID))
			{
				Lacewing::Server::Channel &Channel=Client.ChannelLoopIndex(ID);
				
				// If leaving the channel loop before the end
				Server.EndChannelLoop(ID);
			}

		*/

		// A structure to represent a file being sent to the client
		struct File
		{
			// The name of the file
			const char * Filename;

			// The percentage complete of the file transfer
			int PercentageComplete();
		};

		// Send a file to the client
		LacewingFunction File * SendFile(const char * Filename, int Subchannel=0, __int64 Offset=0, unsigned int Size=-1, const char * VirtualFilename="");

		// Tag (use for whatever you want if you need to)
		void * Tag;
	};

	// The number of clients on the server
	LacewingFunction int ClientCount();

	// Loop through the connected clients
	LacewingFunction int ClientLoop(int ID=0);
	LacewingFunction Client &ClientLoopIndex(int ID);

	// Prematurely end the client loop
	LacewingFunction void EndClientLoop(int ID);

	/*

		for(int ID = 0; ID; ID = Server.ClientLoop(ID))
		{
			Lacewing::Server::Client &Client = Server.ClientLoopIndex(ID);
			
			// If leaving the client loop before the end
			Server.EndClientLoop(ID);
		}

	*/

	// Trigger any waiting handlers (for Lacewing::HandlerMode::Queue or Lacewing::HandlerMode::Suspend)
	LacewingFunction void ProcessHandlers();

	// Available handlers for this class
	typedef Lacewing::Control::Type				(__cdecl * HandlerConnect)									(Lacewing::Server &Server,Lacewing::Server::Client &Client, Lacewing::DenyStatus::Type DenyStatus);
	typedef void								(__cdecl * HandlerDisconnect)								(Lacewing::Server &Server,Lacewing::Server::Client &Client);
	typedef void								(__cdecl * HandlerError)									(Lacewing::Server &Server,Lacewing::Error &);
	typedef void								(__cdecl * HandlerServerMessage)							(Lacewing::Server &Server,Lacewing::Server::Client &Client,Lacewing::TransmissionMode::Type Mode,int Subchannel,char * Packet,int Size,int Type);
	typedef Lacewing::Control::Type				(__cdecl * HandlerChannelMessage)							(Lacewing::Server &Server,Lacewing::Server::Client &Client,Lacewing::Server::Channel &Channel,Lacewing::TransmissionMode::Type Mode,int Subchannel,char * Packet,int Size,int Type);
	typedef Lacewing::Control::Type				(__cdecl * HandlerPeerMessage)								(Lacewing::Server &Server,Lacewing::Server::Client &Client,Lacewing::Server::Channel &Channel,Lacewing::Server::Client &TargetClient,Lacewing::TransmissionMode::Type Mode,int Subchannel,char * Packet,int Size,int Type);
	typedef const char *						(__cdecl * HandlerJoinChannel)								(Lacewing::Server &Server,Lacewing::Server::Client &Client,const char * Name, Lacewing::DenyStatus::Type DenyStatus);
	typedef Lacewing::Control::Type				(__cdecl * HandlerLeaveChannel)								(Lacewing::Server &Server,Lacewing::Server::Client &Client,Lacewing::Server::Channel &Channel, Lacewing::DenyStatus::Type DenyStatus);
	typedef const char *						(__cdecl * HandlerSetName)									(Lacewing::Server &Server,Lacewing::Server::Client &Client,const char * Name, Lacewing::DenyStatus::Type DenyStatus);
	typedef Lacewing::Control::Type				(__cdecl * HandlerWebInterfaceGet)							(Lacewing::Server &Server,Lacewing::Webserver &Webserver,Lacewing::Webserver::Request &Request);
	typedef Lacewing::Control::Type				(__cdecl * HandlerWebInterfacePost)							(Lacewing::Server &Server,Lacewing::Webserver &Webserver,Lacewing::Webserver::Request &Request);
	typedef Lacewing::Control::Type				(__cdecl * HandlerWebInterfaceHead)							(Lacewing::Server &Server,Lacewing::Webserver &Webserver,Lacewing::Webserver::Request &Request);

	// Register handlers
	LacewingFunction void RegisterHandlerConnect(HandlerConnect, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerDisconnect(HandlerDisconnect, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerError(HandlerError, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerServerMessage(HandlerServerMessage, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerChannelMessage(HandlerChannelMessage, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerPeerMessage(HandlerPeerMessage, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerJoinChannel(HandlerJoinChannel, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerLeaveChannel(HandlerLeaveChannel, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerSetName(HandlerSetName, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerWebInterfaceGet(HandlerWebInterfaceGet, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerWebInterfacePost(HandlerWebInterfacePost, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerWebInterfaceHead(HandlerWebInterfaceHead, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);

	// Tag (use for whatever you want if you need to)
	void * Tag;
};

class Ping
{

private:

	// Internal tag (used inside the library)
	void * InternalTag;

public:

	// Constructor and destructor just wrap Initialise and Free
	inline Ping()
	{
		Initialise();
	}

	inline ~Ping()
	{
		Free();
	}
	
	// Initialise the class
	LacewingFunction void Initialise();
	
	// Deinitialise the class
	LacewingFunction void Free();

	// Send one or more ping requests
	LacewingFunction void SendPingRequest(const char * Hostname,int Timeout=250,int RequestCount=1);

	// Trigger any waiting handlers (for Lacewing::HandlerMode::Queue or Lacewing::HandlerMode::Suspend)
	LacewingFunction void ProcessHandlers();

	// Available handlers for this class
	typedef Lacewing::Control::Type			(__cdecl * HandlerPingSuccess)		(Lacewing::Ping &Ping,const char * Hostname,int RequestsLeft,int RTT,int TTL);
	typedef Lacewing::Control::Type			(__cdecl * HandlerPingFailure)		(Lacewing::Ping &Ping,const char * Hostname,int RequestsLeft);
	typedef void							(__cdecl * HandlerError)			(Lacewing::Ping &Ping,Lacewing::Error &);

	// Register handlers
	LacewingFunction void RegisterHandlerPingSuccess(HandlerPingSuccess, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerPingFailure(HandlerPingFailure, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerError(HandlerError, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);

	// Tag (use for whatever you want if you need to)
	void * Tag;
};

class LAN
{

private:

	// Internal tag (used inside the library)
	void * InternalTag;

public:


	// Constructor and destructor just wrap Initialise and Free
	inline LAN()
	{
		Initialise();
	}

	inline ~LAN()
	{
		Free();
	}
	
	// Initialise the class
	LacewingFunction void Initialise();
	
	// Deinitialise the class
	LacewingFunction void Free();

	// Send one or more ping requests
	LacewingFunction void EnumerateDevices();

	// Structure to represent a device found on the LAN
	struct Device
	{
		// Internal tag (used inside the library)
		void * InternalTag;

		// The address of the device as a 4 byte IP address
		unsigned char IP[4];

		// Convert the IP address to a string
		LacewingFunction const char * StringIP();

		// The hostname of the device
		const char * Hostname;
	};

	// Trigger any waiting handlers (for Lacewing::HandlerMode::Queue or Lacewing::HandlerMode::Suspend)
	LacewingFunction void ProcessHandlers();

	// Available handlers for this class
	typedef Lacewing::Control::Type			(__cdecl * HandlerFoundDevice)		(Lacewing::LAN &LAN,Lacewing::LAN::Device &Device);
	typedef void							(__cdecl * HandlerError)			(Lacewing::LAN &LAN,Lacewing::Error &);

	// Register handlers
	LacewingFunction void RegisterHandlerFoundDevice(HandlerFoundDevice, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);
	LacewingFunction void RegisterHandlerError(HandlerError, Lacewing::HandlerMode::Type Mode=Lacewing::HandlerMode::Trigger);

	// Tag (use for whatever you want if you need to)
	void * Tag;
};

// Lacewing namespace
};

// C++ check
#endif

// Include guard
#endif
